/*!
  \file
\verbatim

Copyright (c) 2004, Sylvain Paris and Francois Sillion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    * Neither the name of ARTIS, GRAVIR-IMAG nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

\endverbatim
  This file contains useful functions to compute a convolution between
  two 2D functions. The code is based on FFTW (http://www.fftw.org/), a lot of
  magical numbers are explained on the web site. The system is set to make
  as many optimizations as possible and then it saves them in a file to reuse
  them. Therefore when a new data size is used, it takes some extra time at
  the first run.

  Here is an example:

  \code
  // Array_2D has function members:
  //    x_size(), y_size(), resize(unsigned,unsigned)
  
  Array_2D<double> source,kernel,result;

  // ...
  // Some code to fill source and kernel with data.
  // !!! They must be of the same size.
  // ...

  // Wisdom data will be read and saved in the file "wisdom.fftw".
  // If the file does not exists, it will be created.
  FFT::Support::set_wisdom_file("wisdom.fftw");
  FFT::Support s(source.x_size(),source.y_size());

  // The following line is where the computation actually occurs.
  FFT::convolve(source,kernel,&result,s);

  // result is resized and now contains source*kernel.
  
  // If some convolutions are needed with the same data size,
  // the support s can be reused.
  \endcode

*/

#ifndef __CONVOLUTION_FFT__
#define __CONVOLUTION_FFT__

#include <algorithm>

#include "fftw3.h"

#include "fft/support.h"



namespace FFT{

  
  

  //! Compute h=f*g for f,g and h real functions.
  template<typename Function>
  inline void convolve(const Function& f,
		       const Function& g,
		       Function* const h);

  
  //! Compute h=f*g for f,g and h real functions. Support is given,
  //! goes faster for multiple computations.
  template<typename Function>
  inline void convolve(const Function& f,
		       const Function& g,
		       Function* const h,
		       Support&        support);

  
  //! This function mixes frequency representation for f and space
  //! representation for g to provide a fast process when f is convolved
  //! several times with different g.
  //! Compute h=f*g for f,g and h real functions. Support is given,
  //! goes faster for multiple computations.
  //! f is given through its Fourier transform F in a Support::buffer_type.
  template<typename Function>
  void convolve(const Support::buffer_type& F,
		const Function& g,
		Function* const h,
		Support&        support);

  
  //! Adapt the function so that the point (0,0) is at (x,y) before
  //! the funtion call.
  template<typename Function>
  void move_center(const Function&    f,
		   const unsigned int x,
		   const unsigned int y,
		   Function* const    result);



  


/*
  ###########################################################
  ###########################################################
  ###########################################################
  ##############                               ##############
  ##############  I M P L E M E N T A T I O N  ##############
  ##############                               ##############
  ###########################################################
  ###########################################################
  ###########################################################
*/

  
  template<typename Function>
  void convolve(const Support::buffer_type& F,
		const Function& g,
		Function* const h,
		Support&        support){

    // ################
    // ### FFT of g ###
    // ################

    // Store g data in the fftw structures.    
    support.load_space_data(g);
    
    // Compute the FFT.
    support.space_to_frequency();



    
    // Multiplication "in place" in the frequency space.    
    support.multiply_frequency_data_by_buffer(F,true);


    

    // ###################
    // ### Inverse FFT ###
    // ###################
    
    // Compute inverse FFT transform.
    support.frequency_to_space();

    // Return data into the original structures    
    support.save_space_data(h);
  }


  /*!
    A Function must have members \e x_size() and \e y_size() to get
    the dimensions of the data. It must have an access operator \e (x,y) too.
  */
  template<typename Function>
  void convolve(const Function& f,
		const Function& g,
		Function* const h,
		Support&  support){

    // ################
    // ### FFT of f ###
    // ################

    // We need a temporary buffer.
    Support::buffer_type buffer = support.create_buffer();

    // Store f data in the fftw structures.
    support.load_space_data(f);
    
    // Compute the FFT.
    support.space_to_frequency();
    
    // Copy the result in a safe place.
    support.save_frequency_data_into_buffer(buffer);


    convolve(buffer,g,h,support);
    
    support.destroy_buffer(buffer);
    
//     // ################
//     // ### FFT of g ###
//     // ################

//     // Store g data in the fftw structures.    
//     support.load_space_data(g);
    
//     // Compute the FFT.
//     support.space_to_frequency();



    
//     // Multiplication "in place" in the frequency space.    
//     support.multiply_frequency_data_by_buffer(buffer,true);


    

//     // ###################
//     // ### Inverse FFT ###
//     // ###################
    
//     // Compute inverse FFT transform.
//     support.frequency_to_space();

//     // Return data into the original structures    
//     support.save_space_data(h);

  }

  

  template<typename Function>
  void convolve(const Function& f,
		const Function& g,
		Function* const h){

    Support s(f.x_size(),f.y_size());

    convolve(f,g,h,s);
  }

  
  template<typename Function>
  void move_center(const Function&    f,
		   const unsigned int x_center,
		   const unsigned int y_center,
		   Function* const    result){

    typedef unsigned int size_type;

    const size_type width  = f.x_size();
    const size_type height = f.y_size();

    Function g(width,height);
    
    for(size_type x=0;x<width;x++){
      for(size_type y=0;y<height;y++){
	(*result)(x,y) = f((x + width - x_center) % width,
			   (y + height - y_center) % height);
      }
    }
  }

}

#endif
