/*!
  \file
\verbatim

Copyright (c) 2004, Sylvain Paris and Francois Sillion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    * Neither the name of ARTIS, GRAVIR-IMAG nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

\endverbatim
 *  This file provides tools to achieve to define a complex function
 * from its real and imaginary parts. From a coding point of view,
 * functions are 2D arrays.
 *
 *  All the functions in this file assume that the arrays are used with
 * a FFT::Support.
*/

#ifndef __FFT_FILL__
#define __FFT_FILL__


namespace FFT{

  //! Fill in the real part of the array with function f. Imaginary part is
  //! not changed.
  template <typename Function,typename Complex_data_2D_array>
  void fill_real_part(const Function& f,
		      Complex_data_2D_array* const array);

  //! Fill in the imaginary part of the array with function f. Real part is
  //! not changed.
  template <typename Function,typename Complex_data_2D_array>
  void fill_imaginary_part(const Function& f,
			   Complex_data_2D_array* const array);



  
/*
  ###########################################################
  ###########################################################
  ###########################################################
  ##############                               ##############
  ##############  I M P L E M E N T A T I O N  ##############
  ##############                               ##############
  ###########################################################
  ###########################################################
  ###########################################################
*/



  
  /*!
    Complex_data_2D_array must provide the following functions:
    x_size(), y_size() and a (x,y) access operator.
    Complex_data_2D_array::value_type is assumed to be compatible
    with the standard complex<double> structure.
  */
  template <typename Function,typename Complex_data_2D_array>
  void fill_real_part(const Function& f,
		      Complex_data_2D_array* const array){

    typedef typename Complex_data_2D_array::value_type complex_type;
    typedef float                                      real_type;
    typedef unsigned int                               size_type;

    const size_type width  = array->x_size();
    const size_type height = array->y_size();
  
    const size_type half_width  = width/2;

    for(size_type x=0;x<width;x++){
      for(size_type y=0;y<height;y++){

	const real_type X = static_cast<real_type>(x) -
	  width*((x+half_width)/width);
      
	const real_type Y = static_cast<real_type>(y);
 
	(*array)(x,y) = complex_type(f(X,Y),(*array)(x,y).imag());
      }
    }
  }


  
  /*!
    Complex_data_2D_array must provide the following functions:
    x_size(), y_size() and a (x,y) access operator.
    Complex_data_2D_array::value_type is assumed to be compatible
    with the standard complex<double> structure.
  */
  template <typename Function,typename Complex_data_2D_array>
  void fill_imaginary_part(const Function& f,
			   Complex_data_2D_array* const array){

    typedef typename Complex_data_2D_array::value_type complex_type;
    typedef float                                      real_type;
    typedef unsigned int                               size_type;

    const size_type width  = array->x_size();
    const size_type height = array->y_size();
  
    const size_type half_width  = width/2;

    for(size_type x=0;x<width;x++){
      for(size_type y=0;y<height;y++){

	const real_type X = static_cast<real_type>(x) -
	  width*((x+half_width)/width);
      
	const real_type Y = static_cast<real_type>(y) - 0.5*height;
 
	(*array)(x,y) = complex_type((*array)(x,y).real(),f(X,Y));
      }
    }
  }

}

#endif
